#ident	"$Header: /var/cvsroot/htdb/src/dso/dso_funcs.c,v 1.241 2004/04/30 19:49:37 del Exp $"

/*
	Copyright  1994-2004 David Whittemore
	<del@adjective.com>, all rights reserved.

	Permission to use, copy, modify, and non-commercially distribute
	this software and its documentation for any purpose is hereby granted
	without fee, provided that the above copyright notice appear in all
	copies and that both that copyright notice and this permission notice
	appear in supporting documentation.  No representations are made about
	the suitability of this software for any purpose.  It is provided
	"as is" without express or implied warranty.
 */
#include	"htdb.h"

/*
	welcome to dso_funcs.c

	into this file may be placed functions which will allow
	you to extend the functionality of the HTDB system.

	suppose you desired to add the ability to rotate text
	by 13 characters.  let's call this new function "rot13".

	in HTDB scriptland, you would use the following syntax:

		${rot13(hello world)}

	in this file, you would need to create a new function:

		char *dso_rot13(char *in)

	which would return a pointer to a DYNAMICALLY-ALLOCATED string.

	always: string into a function, string out of a function.

	1) compile.
		and that's about it.
		nice, huh?

	HTDB looks in the conf.htdb datafile for the value of `confDSOs'
	to determine where compiled DSO files may be found.

		example:
			#define	confDSOs
				${_htdb_root}/dso/dso_func.so
 */

extern int evaluate(char *line, double *val);

char	*dso_eatWhitespace(char *in)
{
	char	ret[BUF_LIMIT_SMALL + 1];
	int		indx = 0;
	while (in && *in && indx < BUF_LIMIT_SMALL) {
		if (!isspace((int)*in))
			ret[indx++] = *in;
		in++;
	}
	ret[indx] = '\0';
	return DSO_RETURNS_SUCCESS(ret);
}

char	*dso_ccNumber(char *in)
{
	char	ret[BUF_LIMIT_SMALL + 1];
	int		indx = 0;
	while (in && *in && indx < BUF_LIMIT_SMALL) {
		if (isdigit((int)*in))
			ret[indx++] = *in;
		in++;
	}
	ret[indx] = '\0';
	return DSO_RETURNS_SUCCESS(ret);
}

#define	SYNTAX_ccValid	"usage: \\$\\{ccValid(credit card #)\\}"

char	*dso_ccValid(char *in)
{
	if (ccValid(in))
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_ccValid);
}

/*
	dso_ccType()
		credit card number validation
		uses LUHN Formula (Mod 10) for Validation of Primary Account Number
		see http://www.beachnet.com/~hstiles/cardtype.html
*/
char	*dso_ccType(char *in)
{
	char	*ptr = dso_ccNumber(in),
			*type = "unknown";
	int		length = strlen(ptr);

	if (length == 16 && *ptr == '5' &&
		(*(ptr + 1) == '1' ||
		*(ptr + 1) == '2' ||
		*(ptr + 1) == '3' ||
		*(ptr + 1) == '4' ||
		*(ptr + 1) == '5'))
		type = "mc";
	else if ((length == 13 || length == 16) &&
		*ptr == '4')
		type = "visa";
	else if (length == 15 && *ptr == '3' &&
		(*(ptr + 1) == '4' ||
		*(ptr + 1) == '7'))
		type = "amex";
	else if (length == 14 && *ptr == '3' &&
		((*(ptr + 1) == '0' &&
		(*(ptr + 2) == '0' ||
		*(ptr + 2) == '1' ||
		*(ptr + 2) == '2' ||
		*(ptr + 2) == '3' ||
		*(ptr + 2) == '4' ||
		*(ptr + 2) == '5')) ||
		(*(ptr + 1) == '6' ||
		*(ptr + 1) == '8')))
		type = "diners";
	else if (length == 16 &&
		*ptr == '6' &&
		*(ptr + 1) == '0' &&
		*(ptr + 2) == '1' &&
		*(ptr + 3) == '1')
		type = "discover";
	else if (length == 15 &&
		*ptr == '2' &&
		(
		(*(ptr + 1) == '0' && *(ptr + 2) == '1' && *(ptr + 3) == '4') ||
		(*(ptr + 1) == '1' && *(ptr + 2) == '4' && *(ptr + 3) == '9')
		))
		type = "enroute";
	else if (length == 16 && *ptr == '3')
		type = "jcb";
	else if (length == 15 &&
		(
		(*(ptr + 1) == '2' && *(ptr + 2) == '1' && *(ptr + 3) == '3' &&
			*(ptr + 4) == '1') ||
		(*(ptr + 1) == '1' && *(ptr + 2) == '8' && *(ptr + 3) == '0' &&
			*(ptr + 4) == '0')
		))
		type = "jcb";
	else
		type = "unknown";

	free(ptr);
	return DSO_RETURNS_SUCCESS(type);
}

/*
	string manipulations..
 */
char	*dso_lower(char *source)
{
	char	*ret = strdup(source),
			*hold = &ret[0];

	while (ret && *ret) {
		*(ret) = tolower(*(ret));
		ret++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(hold);
}

char	*dso_upper(char *source)
{
	char	*ret = strdup(source),
			*hold = &ret[0];

	while (ret && *ret) {
		*(ret) = toupper(*(ret));
		ret++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(hold);
}

/* UNTESTED.  Ben, 05-15-2001 */
/*
	dso_index()
		returns a number showing the first occurance of string2 in 
		string 1.  Named according to perl syntax, or to follow
		substr().
*/

#define	SYNTAX_index		"usage: \\$\\{index(string1, string2)\\}"
char	*dso_index(char *in)
{
	char 	**args,
			buf[BUF_LIMIT_SMALL + 1];
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char *strfound;
		strfound = strstr(args[0], args[1]);
		if (strfound) {
			snprintf(buf, BUF_LIMIT_SMALL, "%d", (int) (strfound - args[0]));
			buf[BUF_LIMIT_SMALL] = '\0';
			htdb_deleteArgs(args, num_args);
			return DSO_RETURNS_SUCCESS(buf);
		} else {
			htdb_deleteArgs(args, num_args);
			return DSO_RETURNS_SUCCESS("0");
		}
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_index);
	}
}	

#define	SYNTAX_substr		"usage: \\$\\{substr(start, count, string)\\}"

/*
	dso_substr()
		return a range of characters from within a string
 */
char	*dso_substr(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		int		start = atoi(args[0]),
				range = atoi(args[1]),
				len = strlen(args[2]);
		char	*str = args[2],
				*ret;

		if (start < 0)
			start = 0;

		if (range == 0)
			/*
				substr(4, 0, hello world)
			 */
			range = len;

		if (start > len)
			/*
				asked for past end of string,
				return empty string
			 */
			ret = strdup("");
		else {

			if (range >= 0) {
				/*
					range goes left -> right
				 */
				if (start + range > len)
					/*
						range takes us past the end.
						truncate the range to match string
					 */
					range = len - start;

				ret = strdup(&str[start]);
				ret[range] = '\0';
			} else {
				/*
					range goes right <- left
				 */
				if (start + range + 1 < 0)
					/*
						range takes us past the beginning.
						truncate the range to match string
					 */
					range = 0 - start - 1;

				ret = strdup(&str[start + range + 1]);
				ret[1 - range - 1] = '\0';
			}

		}

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_substr);
	}
}

static char	*_replace(char *source, char *search, char *replace, int case_sensitive)
{
	char * (*sstr) (const char *, const char *);
	spaceData b;
	char *pos, *old;
	int replaceLen = strlen(replace),
		searchLen = strlen(search);

	sstr = case_sensitive ? strstr : strcasestr;

	data_init(&b);

	pos = old = source;
	while ( (pos = sstr(pos, search)) ) {
		thing_append(&b, old, pos - old, BUF_LIMIT_SMALL);
		thing_append(&b, replace, replaceLen, BUF_LIMIT_SMALL);
		pos += searchLen;
		old = pos;
	}
	thing_append(&b, old, (source + strlen(source)) - old, BUF_LIMIT_SMALL);

	return b.value;
}

#define	SYNTAX_replace		"usage: \\$\\{replace(source, searchstring, replacestring)}"

/*
	dso_replace()
		replace all occurances of searchstring within source with replacestring
 */
char	*dso_replace(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char *ret = _replace(args[0], args[1], args[2], 1);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else if (num_args == 2) {
		char *ret = _replace(args[0], args[1], "", 1);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_replace);
	}
}

#define	SYNTAX_replacei		"usage: \\$\\{replacei(source, searchstring, replacestring)}"

/*
	dso_replacei()
		replace all occurances of searchstring within source with replacestring
 */
char	*dso_replacei(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char *ret = _replace(args[0], args[1], args[2], 0);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else if (num_args == 2) {
		char *ret = _replace(args[0], args[1], "", 0);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_replacei);
	}
}

char	*dso_pretty(char *ugly)
{
	return DSO_RETURNS_SUCCESS_ASIS(pretty(ugly));
}

char	*dso_pretty_state(char *ugly)
{
	return DSO_RETURNS_SUCCESS_ASIS(prettyState(ugly));
}

char	*dso_pretty_date(char *ugly)
{
	char	pretty[BUF_LIMIT_SMALL + 1];
	time_t	t = atol(ugly);
	sstrncpy(pretty, ctime(&t));
	pretty[24] = '\0';
	return DSO_RETURNS_SUCCESS(pretty);
}

char	*dso_pretty_time(char *ugly)
{
	double	min;
	int		hour;
	char	pretty[BUF_LIMIT_SMALL + 1],
			*c;
	if ((c = strchr(ugly, '.'))) {
		hour = atoi(ugly);
		min = atof(c);
	} else {
		hour = atoi(ugly);
		min = 0.0;
	}
	c = (hour < 12 || hour > 23) ? "AM" : "PM";
	if (hour == 0) hour = 12;
	if (hour > 12) hour -=12;
	if (min == 0)
		snprintf(pretty, BUF_LIMIT_SMALL, "%d %s",
			hour, c);
	else
		snprintf(pretty, BUF_LIMIT_SMALL, "%d:%2d %s",
			hour, (int)(min * 60), c);
	pretty[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(pretty);
}

char	*dso_encrypt(char *source)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_encryptor("pe1", source));
}

char	*dso_be1(char *source)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_encryptor("be1", source));
}

char	*dso_se1(char *source)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_encryptor("se1", source));
}

char	*dso_pe1(char *source)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_encryptor("pe1", source));
}

#define	SYNTAX_indexOf	"usage: \\$\\{indexOf(table, field, value)\\}"

char	*dso_indexOf(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	*table = args[0],
				*field = args[1],
				*value = args[2];
		int	status = myIndexOf(table, field, value);

		htdb_deleteArgs(args, num_args);

		if (status) {
			char	ret[BUF_LIMIT_SMALL + 1];
			snprintf(ret, BUF_LIMIT_SMALL, "%d", status);
			return DSO_RETURNS_SUCCESS(ret);
		} else
			return DSO_RETURNS_SUCCESS_CHALLENGED("");
	} else {
		htdb_deleteArgs(args, num_args);
	}
	return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_indexOf);
}

#define	SYNTAX_idOf	"usage: \\$\\{idOf(table, field, value)\\}"

char	*dso_idOf(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	*table = args[0],
				*field = args[1],
				*value = args[2];
		int	this_index,
			found = 1;

		for (this_index = 1 ; found == 1 ; this_index++) {
			char	*lookup = htdb_getobjname(table, this_index, field);

			if (htdb_checkval(lookup)) {
				if (htdb_equal(lookup, value)) {
					char	ret[BUF_LIMIT_SMALL + 1];

					snprintf(ret, BUF_LIMIT_SMALL, "%s[%d]->%s_id",
						table, this_index, table);
					ret[BUF_LIMIT_SMALL] = '\0';

					htdb_deleteArgs(args, num_args);
					return DSO_RETURNS_SUCCESS(htdb_getval(ret));
				}
			} else
				found = 0;
		}
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_idOf);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_idOf);
	}
}

#define	SYNTAX_trunc	"usage: \\$\\{trunc(maxlen, string)\\}"

char	*dso_trunc(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		int		len = atoi(args[0]);
		/*  take out anything that could screw up our trunc 
		 	Ok to be unsafe here because we immediately re-filter. */
		char	*compact = htdb_unfilter_unsafe(args[1]),
				*out = (char *)malloc(sizeof(char) * (len + 4)),
				*full;
		
		strncpy(out, compact, len); out[len] = '\0';

		if (strlen(compact) > len)
			strcat(out, "..");
		
		full = filterProgramArguments(out);
		free(out);
		free(compact);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(full);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS("");
	}
}

#define	SYNTAX_safetrunc 	"usage: \\$\\{safetrunc(maxlen | string)\\}"

char	*dso_safetrunc(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, "|", &args, 2);

	if (num_args == 2) {
		char	*ret = html_safe_trunc(args[1], atoi(args[0]));
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else  {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_safetrunc);
	}
}

char	*dso_formSafe(char *in)
{
	char	*out = strdup(in),
			*ptr = &out[0];
	while (ptr && *ptr) {
		if (*ptr == '\n' || *ptr == '\r')
			*ptr = ' ';
		else if (*ptr == '"')
			*ptr = '\'';
		ptr++;
	}
	return DSO_RETURNS_SUCCESS_ASIS(out);
}

char	*dso_abs(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];
	snprintf(out, BUF_LIMIT_SMALL, "%ld", labs(atol(in)));
	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(out);
}

#define	SYNTAX_prettyWrap	"usage: \\$\\{prettyWrap([margin, ]string)\\}"

char	*dso_prettyWrap(char *in)
{
	int	indx = 0,
		margin = 0;
	char	*comma = strchr(in, ','),
			*out = (char *)malloc(sizeof(char) * ((strlen(in) * 2) + 1)),
			*hold = &out[0];

	if (comma && *comma) {
		*comma = '\0';
		if (atoi(in) > 0) {
			margin = atoi(in);
			in = comma + 1;
		} else
			*comma = ',';

	}
	while (in && *in) {
		if (indx > 0 && *in == '\n') {
			*(out++) = ' ';
		} else if (*in == '-' && (in - 1 && *(in - 1) == '\n')) {
			indx = 0;
			*(out++) = '\n';
		} else {
			if (++indx > margin && isspace((int)*in)) {
				*(out++) = '\n';
				indx = 0;
			} else {
				*(out++) = *in;
			}
		}
		in++;
	}
	*out = '\0';
	return DSO_RETURNS_SUCCESS_ASIS(hold);
}

/*
	dso_encode()
		URL encoding
 */
char	*dso_encode(char *source)
{
	return DSO_RETURNS_SUCCESS_ASIS(encode(source));
}

#define	SYNTAX_isql		"isql: ${_error_query}"

/*
	dso_isql()
		makes a SQL call and makes callbacks for each row of the result
		set to the resource named li`prefix' and build-ups a result set
		named lo`prefix'
 */
char	*dso_isql(char *source)
{
	char	*prefix,
		*colon = strchr(source, ':');

	if (colon && (colon - source) < 20) {
		/*
			user-selected prefix for results
			(only if the colon is in the first 20 characters)
		 */
		prefix = source;
		*colon = '\0';
		source = colon + 1;
	} else
		/*
			default
		 */
		prefix = "sql";

	if (dbProcessByRow(source, prefix) < 0)
		/*
			error
		 */
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_isql);
	else
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
}

#define	SYNTAX_sql		"usage: sql(table:SQL) - bad SQL: \"${_error_query}\""

/*
	dso_sql()
		makes a SQL call and stores the result set in `prefix'-named namespace
 */
char	*dso_sql(char *source)
{
	char	*prefix,
		*colon = strchr(source, ':');

	if (colon && (colon - source) < 20) {
		/*
			user-selected prefix for results
			(only if the colon is in the first 20 characters)
		 */
		prefix = source;
		*colon = '\0';
		source = colon + 1;
	} else
		/*
			default
		 */
		prefix = "sql";

	if (dbGetRowByKey(prefix, source) < 0)
		/*
			error
		 */
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_sql);
	else
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
}

#define SYNTAX_system	"usage: \\$\\{system(command)\\}"

/*
	dso_system()
		makes a system call as whatever user the process is running as.
		begeesus.  i can't believe i'm including this..
		we try to do *some* damage control by not allowing root processes.
 */
char	*dso_system(char *source)
{
	if (BOOL_TRUE == amRoot())
		return DSO_RETURNS_SUCCESS_CHALLENGED(ARE_YOU_INSANE);
	else {
		return DSO_RETURNS_SUCCESS_ASIS(htdb_system(source));
	}
}

/*
	dso_random()
		where `token' is a string representation of
		a number, return a random number modulo
		the integer value of `token'.
 */
char	*dso_random(char *token)
{
	char	ret[BUF_LIMIT_SMALL + 1],
			*expanded = htdb_expand(token);

	snprintf(ret, BUF_LIMIT_SMALL, "%ld", htdb_random(atol(expanded)));
	ret[BUF_LIMIT_SMALL] = '\0';
	free(expanded);

	return DSO_RETURNS_SUCCESS(ret);
}

/*
	dso_strlen()
		return (a string representation)
		of the number of characters in `token'.
 */
char	*dso_strlen(char *token)
{
	char	ret[BUF_LIMIT_SMALL + 1],
			*expanded = htdb_expand(token);

	snprintf(ret, BUF_LIMIT_SMALL, "%d",
		(int)strlen(expanded));
	ret[BUF_LIMIT_SMALL] = '\0';
	free(expanded);
	return DSO_RETURNS_SUCCESS(ret);
}

/*
	dso_space2underscore()
		turns spaces into underscores
 */
char	*dso_space2underscore(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(space2underscore(in));
}

/*
	dso_nonalnum2underscore()
		turns things not alnums into underscores
 */
char	*dso_nonalnum2underscore(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(makeFileSystemSafe(strdup(in)));
}

/*
	dso_int()
		returns the integer value of the string
 */
char	*dso_int(char *in)
{
	char	buh[BUF_LIMIT_SMALL + 1];
	snprintf(buh, BUF_LIMIT_SMALL, "%ld", atol(in));
	buh[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(buh);
}

#define	SYNTAX_goLimited		"usage: \\$\\{goLimited(count, url)\\}"

/*
	dso_goLimited()
		inserts a row into the go table
		with limiting characteristics
 */
char	*dso_goLimited(char *in)
{
	char	*comma = strchr(in, ',');
	if (comma && *comma) {
		char	*key = genGoKey(),
				*server = htdb_script("serverHTTP"),
				*ptr,
				*dest;
		int	maxcount;

		*comma = '\0';

		if ( !key )
			return DSO_RETURNS_SUCCESS_CHALLENGED("");

		maxcount = atoi(in);

		dest = strstr(comma + 1, "http");
		ptr = strstr(dest, server);

		if (ptr && *ptr)
			dest += strlen(server);

		dbInsert("insert into go set "
			"go_id=NULL, "
			"gokey='%s', "
			"dest='%s', "
			"count=0, "
			"maxcount=%d, "
			"lastused=now(), "
			"startdate='', "
			"stopdate='', "
			"isdeleted='F'",
			key, dest, maxcount);

		free(server);

		return DSO_RETURNS_SUCCESS_ASIS(key);
	} else
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_goLimited);
}

/*
	dso_goUnlimited()
		inserts a row into the go table
		with unlimited characteristics
 */
char	*dso_goUnlimited(char *in)
{
	char	*key = genGoKey(),
			*server = htdb_script("serverHTTP"),
			*ptr = strstr(in, server);

	if ( !key )
		return DSO_RETURNS_SUCCESS_CHALLENGED("");

	if (ptr && *ptr)
		in += strlen(server);

	dbInsert("insert into go set "
		"go_id=NULL, "
		"gokey='%s', "
		"dest='%s', "
		"count=0, "
		"maxcount=0, "
		"lastused=now(), "
		"startdate='', "
		"stopdate='', "
		"isdeleted='F'",
		key, in);

	free(server);

	return DSO_RETURNS_SUCCESS_ASIS(key);
}

/*
	dso_dbSafe()
		literals single quote characters
		(should probably also literal double quote characters)
 */
char	*dso_dbSafe(char *in)
{
	char	*out = (char *)malloc(sizeof(char) * ((2 * strlen(in)) + 1));
	int	indx = 0;

	while (in && *in) {
		if (*in == '\'') {
			out[indx++] = '\\';
			out[indx++] = *in;
		} else if (isalnum((int)*in) || *in == ' ' || *in == '.' || *in == '@' || *in == '/')
			out[indx++] = *in;
		in++;
	}

	out[indx] = '\0';
	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_log_debug()
		show default-level debugging
 */
char	*dso_log_debug(char *in)
{
	char	*comma = strchr(in, ','),
			*out = (char *)NULL;

	if (comma && *comma) {
		*comma = '\0';
		debug_it(DEBUG_LEVEL_DEBUG, in, comma + 1, &out);
	} else
		debug_it(DEBUG_LEVEL_DEBUG, in, NULL, &out);

	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_log_debug()
		show info-level debugging
 */
char	*dso_log_info(char *in)
{
	char	*comma = strchr(in, ','),
			*out = (char *)NULL;

	if (comma && *comma) {
		*comma = '\0';
		debug_it(DEBUG_LEVEL_INFO, in, comma + 1, &out);
	} else
		debug_it(DEBUG_LEVEL_INFO, in, NULL, &out);

	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_log_warning()
		show warning-level debugging
 */
char	*dso_log_warning(char *in)
{
	char	*comma = strchr(in, ','),
			*out = (char *)NULL;

	if (comma && *comma) {
		*comma = '\0';
		debug_it(DEBUG_LEVEL_WARNING, in, comma + 1, &out);
	} else
		debug_it(DEBUG_LEVEL_WARNING, in, NULL, &out);

	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_log_error()
		show error-level debugging
 */
char	*dso_log_error(char *in)
{
	char	*comma = strchr(in, ','),
			*out = (char *)NULL;

	if (comma && *comma) {
		*comma = '\0';
		debug_it(DEBUG_LEVEL_ERROR, in, comma + 1, &out);
	} else
		debug_it(DEBUG_LEVEL_ERROR, in, NULL, &out);

	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_log_fatal()
		show fatal-level debugging
 */
char	*dso_log_fatal(char *in)
{
	char	*comma = strchr(in, ','),
			*out = (char *)NULL;

	if (comma && *comma) {
		*comma = '\0';
		debug_it(DEBUG_LEVEL_FATAL, in, comma + 1, &out);
	} else
		debug_it(DEBUG_LEVEL_FATAL, in, NULL, &out);

	return DSO_RETURNS_SUCCESS_ASIS(out);
}

#define	SYNTAX_include	"usage: \\$\\{include(HTDB resource file in ${_htdb_root}/[language]/[branding])\\}"

/*
	dso_include()
		include another HTDB file in-place
 */
char	*dso_include(char *in)
{
	if (htdb_include(in))
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_include);
}

#define	SYNTAX_fileExists	"fileExists: no such file"

/*
	dso_fileExists()
		make a filesystem call to see if file exists
 */
char	*dso_fileExists(char *in)
{
	struct	stat    buf;
	if (stat(in, &buf) == 0)
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fileExists);
}

/*
	dso_shh()
		return an empty string
 */
char	*dso_shh(char *in)
{
	return DSO_RETURNS_SUCCESS("");
}

/*
	dso_deleteSessionData()
		deletes all sessiondata table rows for the current session_id
 */
char	*dso_deleteSessionData(char *in)
{
	int_t	session_id = htdb_getint("_session_int");
	deleteSessionData(session_id, in);
	return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
}

#define	SYNTAX_storeSessionData	"usage: \\$\\{storeSessionData(name, value)\\}"

char	*dso_storeSessionData(char *in)
{
	int_t	session_id = htdb_getint("_session_int");
	char	*colon = strchr(in, ',');
	if (colon && *colon) {
		*colon = '\0';
		storeSessionData(session_id, in, colon + 1);
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	} else
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_storeSessionData);
}

char	*dso_fetchSessionData(char *in)
{
	int_t	session_id = htdb_getint("_session_int");
	fetchSessionData(session_id, in);
	if (in && *in)
		return DSO_RETURNS_SUCCESS(htdb_getval("sessiondata->value"));
	else
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
}

/*
	dso_deleteUserData()
		deletes all userdata table rows for the current user_id
 */
char	*dso_deleteUserData(char *in)
{
	int_t	user_id = htdb_getint("user->user_id");
	if (user_id > 0) {
		deleteUserData(user_id, in);
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	} else
		return DSO_RETURNS_SUCCESS("");
}

#define	SYNTAX_storeUserData	"usage: \\$\\{storeUserData(name, value)\\}"

char	*dso_storeUserData(char *in)
{
	int_t	user_id = htdb_getint("user->user_id");
	if (user_id > 0) {
		char	*colon = strchr(in, ',');
		if (colon && *colon) {
			*colon = '\0';
			storeUserData(user_id, in, colon + 1);
			return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
		} else
			return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_storeUserData);
	} else
		return DSO_RETURNS_SUCCESS("");
}

char	*dso_fetchUserData(char *in)
{
	int_t	user_id = htdb_getint("user->user_id");
	if (user_id > 0) {
		fetchUserData(user_id, in);
		if (in && *in)
			return DSO_RETURNS_SUCCESS(htdb_getval("userdata->value"));
		else
			return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	} else
		return DSO_RETURNS_SUCCESS("");
}

/*
	dso_reverse()
		reverses a string
 */
char	*dso_reverse(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(reverseString(in));
}

/*
	dso_stripHTML()
		turns angle brackets into HTML-printable equivalents
 */
char	*dso_stripHTML(char *in)
{
	int	in_tag = 0;

	spaceData	b;

	data_init(&b);

	while (in && *in) {
		if (*in == '<') {
			if ((in - 1) && *(in - 1) == '\\') {
				thing_appendChar(&b, *in, BUF_LIMIT_BIG);
				in_tag = 1;
			} else {
				/*
					all others get converted into
					a displayable equivalent
				 */
				thing_append(&b, "&lt;", 4, BUF_LIMIT_BIG);
			}
		} else if (*in == '>') {
			if (in_tag) {
				thing_appendChar(&b, *in, BUF_LIMIT_BIG);
				in_tag = 0;
			} else {
				if ((in - 1) && *(in - 1) == '\\')
					thing_appendChar(&b, *in, BUF_LIMIT_BIG);
				else
					/*
						all others get converted into
						a displayable equivalent
					 */
					thing_append(&b, "&gt;", 4, BUF_LIMIT_BIG);
			}
		} else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_filterHTML()
		only allows certain HTML tags through, converts
		others into HTML-printable versions.
 */
char	*dso_filterHTML(char *in)
{
	int	in_tag = 0,
		cnt = 0;

	spaceData	b;

	data_init(&b);

	while (in && *in) {
		if (isspace((int)*in))
			cnt = 0;

		if (++cnt < 35) {
			if (*in == '<') {
				if ((in - 1) && *(in - 1) == '\\')
					thing_appendChar(&b, *in, BUF_LIMIT_BIG);
				else {
					char	*gottag = strdup("");
					/*
						we will only let the following subset
						of HTML pass by..
					 */
					if (isTag("p", in, &gottag) ||
						isTag("font", in, &gottag) ||
						isTag("/font", in, &gottag) ||
						isTag("strong", in, &gottag) ||
						isTag("/strong", in, &gottag) ||
						isTag("em", in, &gottag) ||
						isTag("/em", in, &gottag) ||
						isTag("h1", in, &gottag) ||
						isTag("/h1", in, &gottag) ||
						isTag("h2", in, &gottag) ||
						isTag("/h2", in, &gottag) ||
						isTag("h3", in, &gottag) ||
						isTag("/h3", in, &gottag) ||
						isTag("u", in, &gottag) ||
						isTag("/u", in, &gottag) ||
						isTag("/p", in, &gottag) ||
						isTag("br", in, &gottag) ||
						isTag("/b", in, &gottag) ||
						isTag("b", in, &gottag) ||
						isTag("ol", in, &gottag) ||
						isTag("/ol", in, &gottag) ||
						isTag("ul", in, &gottag) ||
						isTag("/ul", in, &gottag) ||
						isTag("li", in, &gottag) ||
						isTag("/li", in, &gottag) ||
						isTag("/i", in, &gottag) ||
						isTag("i", in, &gottag) ||
						isTag("hr", in, &gottag)) {
						int	len = strlen(gottag);
						in_tag = 1;
						thing_appendChar(&b, '<', BUF_LIMIT_BIG);
						thing_append(&b, gottag, len, BUF_LIMIT_BIG);
						in += len;
					} else {
						/*
							all others get converted into
							a displayable equivalent
						 */
						thing_append(&b, "&lt;", 4, BUF_LIMIT_BIG);
					}
					free(gottag);
				}
			} else if (*in == '>') {
				if (in_tag) {
					thing_appendChar(&b, *in, BUF_LIMIT_BIG);
					in_tag = 0;
				} else {
					if ((in - 1) && *(in - 1) == '\\')
						thing_appendChar(&b, *in, BUF_LIMIT_BIG);
					else {
						thing_append(&b, "&gt;", 4, BUF_LIMIT_BIG);
					}
				}
			} else
				thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		}
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_smartHREF()
		given a string with at least one `.', add "http://" if necessary
		and output the result
 */
char	*dso_smartHREF(char *in)
{
	if (strchr(in, '.') && !strchr(in, '@')) {
		/*
			can't be a URL without at least a fricking DOT, now can it?
		 */
		char	*url = (char *)malloc(sizeof(char) * (strlen(in) + 30));
		if (strncasecmp(in, "http://", 7) == 0)
			strcpy(url, in);
		else
			sprintf(url, "http://%s", in);
		
		return DSO_RETURNS_SUCCESS_ASIS(url);
	} else
		return DSO_RETURNS_SUCCESS(in);
}

/*
	dso_smartURL()
		turn the incoming string into an HREF-able item.
		trys to work with URLS and mailtos.
 */
char	*dso_smartURL(char *in)
{
	if (strchr(in, '.') && !strchr(in, '@')) {
		/*
			can't be a URL without at least a fricking DOT, now can it?
		 */
		char	*url = (char *)malloc(sizeof(char) * (strlen(in) + 30));
		if (strncasecmp(in, "http://", 7) == 0)
			strcpy(url, in);
		else
			sprintf(url, "http://%s", in);
		{
			char	*out = (char *)malloc(sizeof(char) * ((strlen(url) * 2) + 30)),
					*displayurl = strdup(url);
			if (strlen(displayurl) > 50) {
				displayurl[47] = '\0';
				strcat(displayurl, "...");
			}
			sprintf(out, "<a href=\"%s\" target=\"new\">%s</a>", url, displayurl);
			free(url);
			free(displayurl);
			return DSO_RETURNS_SUCCESS_ASIS(out);
		}
	} else if (BOOL_TRUE == isValidEmailAddress(in)) {
		/*
			looks like an email address.
			make a mailto link
		 */
		char	*out = (char *)malloc(sizeof(char) * ((strlen(in) * 2) + 30)),
				*displayurl = strdup(in);
		if (strlen(displayurl) > 50) {
			displayurl[47] = '\0';
			strcat(displayurl, "...");
		}
		sprintf(out, "<a href=\"mailto:%s\">%s</a>", in, displayurl);
		free(displayurl);
		return DSO_RETURNS_SUCCESS_ASIS(out);
	} else
		return DSO_RETURNS_SUCCESS(in);
}

/*
	dso_padZero()
		if a number is < 10, then represent it as "03"
 */
char	*dso_padZero(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];

	snprintf(out, BUF_LIMIT_SMALL, "%02d", atoi(in));
	out[BUF_LIMIT_SMALL] = '\0';

	return DSO_RETURNS_SUCCESS(out);
}

#define	PADSPACE_USAGE_STRING "usage: pad(string, padCount, [l(eft)|r(ight)])"


char	*dso_padSpace(char *in)
{
	char	**args, *out;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);
	int		i, padNeeded, pad;
	int 	padLeft = 0;

	if ( num_args < 2 || num_args > 3) {
		htdb_deleteArgs(args, num_args); 
		return DSO_RETURNS_SUCCESS_CHALLENGED(PADSPACE_USAGE_STRING);
	}

	if ( num_args == 3 ) {
		if ( args[2] && toupper((int) *(args[2])) == 'L' )
			padLeft = 1;
	}

	pad = atoi(args[1]);

	padNeeded = pad - strlen(args[0]);

	if (padNeeded <= 0) {
		out = strdup(args[0]);
	} else {
		out = malloc(pad + 1);
		if ( !out )  {
			htdb_deleteArgs(args, num_args);
			return DSO_RETURNS_SUCCESS_CHALLENGED("out of memory!");
		} 
		
		if ( padLeft ) {
			for(i=0; i < padNeeded; i++) 
				out[i] = ' ';
			strcpy(out + i, args[0]);
		} else {
			strcpy(out, args[0]);
			for(i = strlen(args[0]); i < padNeeded + strlen(args[0]); i++) 
				out[i] = ' ';
			out[i] = '\0';
		}
	}
		
	htdb_deleteArgs(args, num_args); 
	return DSO_RETURNS_SUCCESS_ASIS(out);
}
	
		

/*
	dso_prettySeconds()
			1 year 212 days 19:53:42
			254 days 23:55:00
			10:23
			23 minutes
			10 seconds
		`in' assumed to be in SECONDS
 */
char	*dso_prettySeconds(char *in)
{
	return DSO_RETURNS_SUCCESS(prettySeconds(atol(in)));
}

/*
	dso_prettySecondsTerse()
		outputs of the form:
			1 year, 5 days
			56 days, 10 hours
			10 hours, 23 minutes
			23 minutes, 10 seconds
			10 seconds
		`in' assumed to be in SECONDS
 */
char	*dso_prettySecondsTerse(char *in)
{
	return DSO_RETURNS_SUCCESS(prettySecondsTerse(atol(in)));
}

/*
	dso_prettySecondsTerser()
		outputs of the form:
			1 year
			56 days
			10 hours
			23 minutes
			10 seconds
		`in' assumed to be in SECONDS
 */
char	*dso_prettySecondsTerser(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];
	long	tm = atol(in);
	int		secs = tm % 60,
			mins = (tm / 60) % 60,
			hours = (tm / 3600) % 24,
			days = ((tm / 3600) / 24) % 365,
			years = (tm / 31536000);

	if (years > 0) {
		snprintf(out, BUF_LIMIT_SMALL, "%d year%s",
			years, (years > 1) ? "s" : "");
		return DSO_RETURNS_SUCCESS(out);
	}

	if (days > 0) {
		snprintf(out, BUF_LIMIT_SMALL, "%d day%s",
			days, (days > 1) ? "s" : "");
		return DSO_RETURNS_SUCCESS(out);
	}

	if (hours > 0) {
		snprintf(out, BUF_LIMIT_SMALL, "%d hour%s",
			hours,
			(hours > 1 ? "s" : ""));
		return DSO_RETURNS_SUCCESS(out);
	}

	if (mins > 0) {
		snprintf(out, BUF_LIMIT_SMALL, "%d minute%s",
			mins,
			(mins > 1 ? "s" : ""));
		return DSO_RETURNS_SUCCESS(out);
	} else if (secs > 0) {
		snprintf(out, BUF_LIMIT_SMALL, "%d second%s",
			secs,
			(secs > 1 ? "s" : ""));
		return DSO_RETURNS_SUCCESS(out);
	} else
		strcpy(out, "now");

	return DSO_RETURNS_SUCCESS(out);
}

/*
	dso_truncEmail()
		filters `XX@YYY.COM' -> `XXX@...'
 */
char	*dso_truncEmail(char *in)
{
	char	*dp = strdup(in),
			*at = strchr(dp, '@');

	if (at && *at) {
		*(at + 1) = '\0';
		strcat(dp, "...");
	}

	return DSO_RETURNS_SUCCESS_ASIS(dp);
}

/*
	dso_javascriptDocWrite()
		for each newline-terminated line in the input, output:
			document.write("XXXX");
		where `XXXX' is a line of the input
 */
char	*dso_javascriptDocWrite(char *in)
{
	spaceData	b;

	data_init(&b);

	thing_append(&b, "document.write(\"", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	while (in && *in) {
		if (*in == '\"')
			thing_appendChar(&b, '\'', BUF_LIMIT_BIG);
		else if (*in == '\n' || *in == '\r')
			thing_append(&b, " \\\\n\");\ndocument.write(\"", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in != '\t')
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}
	thing_append(&b, "\");\n", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

#define	PLURALIZE_USAGE_STRING	"usage: pluralize(x, singular, plural)\\}"

/*
	dso_pluralize()
		if the first argument is singular, then output the
		second argument modified, else output the 3rd argument modified
 */
char	*dso_pluralize(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	out[BUF_LIMIT_SMALL + 1];
		int		num = atoi(args[0]);

		if (num == 1)
			snprintf(out, BUF_LIMIT_SMALL, "%d %s", num, args[1]);
		else if (num == 0)
			snprintf(out, BUF_LIMIT_SMALL, "no %s", args[2]);
		else
			snprintf(out, BUF_LIMIT_SMALL, "%d %s", num, args[2]);
		out[BUF_LIMIT_SMALL] = '\0';

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(out);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(PLURALIZE_USAGE_STRING);
	}
}

/*
	dso_unfilter()
		converts HTML escape sequences back to quotes, backslashes, etc 
 */
char	*dso_unfilter(char *jin)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_unfilter(jin)); 
}

/*
	dso_unfilter_unsafe()
		converts HTML escape sequences back to quotes, backslashes, etc 
 */
char	*dso_unfilter_unsafe(char *jin)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_unfilter_unsafe(jin)); 
}

/*
	dso_unfilter_noDoubleQuotes()
		like unfilter except turns double quotes into single quotes.
 */
char	*dso_unfilter_noDoubleQuotes(char *jin)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_unfilter_nodoublequotes(jin)); 
}

/*
	dso_hex2base10()
		given a hex number as input, output a base-10 number
 */
char	*dso_hex2base10(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];
	snprintf(out, BUF_LIMIT_SMALL, "%ld", strtol(in, (char **)NULL, 16));
	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(out);
}

/*
	dso_hex2base10()
		returns hex of the input
 */
char	*dso_num2hex(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];
	snprintf(out, BUF_LIMIT_SMALL, "%x", atoi(in));
	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(out);
}

/*
	dso_prettyNumber()
		1        -> 1
		100      -> 100
		1000     -> 1,000
		1234567  -> 1,234,567
		etc
 */
char	*dso_prettyNumber(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_prettyNumber(in));
}

char	*dso_month2num(char *in)
{
	char	*ret = htdb_month2num(in);

	if (ret && *ret)
		return DSO_RETURNS_SUCCESS(ret);

	return DSO_RETURNS_SUCCESS_CHALLENGED(in);
}

char	*dso_num2mon(char *in)
{
	char	*ret = htdb_num2mon(in);
	
	if (ret && *ret)
		return DSO_RETURNS_SUCCESS(ret);

	return DSO_RETURNS_SUCCESS_CHALLENGED(in);
}

char	*dso_num2month(char *in)
{
	char	*ret = htdb_num2month(in);
	
	if (ret && *ret)
		return DSO_RETURNS_SUCCESS(ret);

	return DSO_RETURNS_SUCCESS_CHALLENGED(in);
}

/*
	dso_fetchNextMonth()
		given a number between 1 and 12, return the integer value
		of the next month
			1 -> 2
			12 -> 1
 */
char	*dso_fetchNextMonth(char *in)
{
	char	*ret = htdb_fetchNextMonth(in);

	if (ret && *ret)
		return DSO_RETURNS_SUCCESS(ret);

	return DSO_RETURNS_SUCCESS_CHALLENGED(in);
}

/*
	dso_fetchPrevMonth()
		given a number between 1 and 12, return the integer value
		of the previous month
			1 -> 12
			2 -> 1
 */
char	*dso_fetchPrevMonth(char *in)
{
	char	*ret = htdb_fetchPrevMonth(in);

	if (ret && *ret)
		return DSO_RETURNS_SUCCESS(ret);

	return DSO_RETURNS_SUCCESS_CHALLENGED(in);
}

/*
	dso_checkFeedback()
		return number of comments left for the specified domain_id
		and resource (page) value
 */
char	*dso_checkFeedback(char *in)
{
	char	*colon = strchr(in, ':'),
		out[512];
	int_t	domain_id = 0;

	if (colon && *colon) {
		*colon = '\0';
		if (atoi(in) > 0) {
			domain_id = atol(in);
			in = colon + 1;
		} else
			*colon = ':';

		dbGetRowByKey("fdbk",
			"select count(feedback_id) as cnt from feedback where uri = '%s' and "
			"domain_id = %ld", in, domain_id);

		sprintf(out, "%s comment%s",
			htdb_getint("fdbk->cnt") > 0 ? htdb_getval("fdbk->cnt") : "no",
			htdb_getint("fdbk->cnt") == 1 ? "" : "s");
	} else
		strcpy(out, "no comments");

	return DSO_RETURNS_SUCCESS(out);
}

#define	SYNTAX_define	"usage: \\$\\{define([name], [value])\\}"

/*
	dso_define()
		inline equivalent of #define
 */
char	*dso_define(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 1 || num_args == 2) {
		char	*sparen,
			*eparen;
		if ((sparen = strchr(args[0], '(')) &&
			(eparen = strchr(args[0], ')'))) {
			/*
				things get trippy..
				if the `name' argument contains parenthesis,
				then we assume we are being asked to declare
				a script function dynamically.
				self-creating code, anyone?
			 */
			*sparen = '\0';
			*eparen = '\0';
			htdb_setfunc(args[0], sparen + 1, (num_args == 1 ? "" : args[1]));
		} else {
			/*
				otherwise, this is a vanilla definition.
			 */
			htdb_setval(args[0], (num_args == 1 ? "" : args[1]));
		}
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS("");
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(SYNTAX_define);
	}
}

#define	SYNTAX_resolve	"usage: \\$\\{resolve([variable], [default])\\}"

char	*dso_resolve(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 1 || num_args == 2) {
		char	*val = args[0],
				*def = (num_args == 1 ? "" : args[1]),
				*ret;

		if (htdb_defined(val)) {
			ret = htdb_script(val);
		} else {
			ret = strdup(def);
		}
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(SYNTAX_define);
	}
}

/*
	dso_fetchTimestamp()
		return unix timestamp w/microseconds
 */
char	*dso_fetchTimestamp(char *in)
{
	struct timeval  now;
	char	buf[BUF_LIMIT_SMALL + 1];

	gettimeofday(&now, (void *)NULL);

	snprintf(buf, BUF_LIMIT_SMALL, "%ld.%06ld",
		now.tv_sec, now.tv_usec);
	buf[BUF_LIMIT_SMALL] = '\0';

	return DSO_RETURNS_SUCCESS(buf);
}

#define	SYNTAX_linkInternal	"usage: \\$\\{linkInternal(href [, text])\\}"

/*
	dso_linkInternal()
		turn the incoming value(s) into a simple HREF
 */
char	*dso_linkInternal(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*href = args[0],
			*text = args[1],
			*buf = (char *)malloc(sizeof(char) *
				(strlen(href) + strlen(text) + 100));
		if (strstr(href, "@"))
			sprintf(buf, "<a href=\"mailto:%s\">%s</a>", href, text);
		else
			sprintf(buf, "<a href=\"%s\">%s</a>", href, text);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(buf);
	} if (num_args == 1) {
		char	*href = args[0],
			*text = args[0],
			*buf = (char *)malloc(sizeof(char) *
				(strlen(href) + strlen(text) + 100));
		if (strstr(href, "@"))
			sprintf(buf, "<a href=\"mailto:%s\">%s</a>", href, text);
		else
			sprintf(buf, "<a href=\"%s\">%s</a>", href, text);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(buf);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(SYNTAX_linkInternal);
	}
}

#define	SYNTAX_linkExternal	"usage: \\$\\{linkExternal(href [, text])\\}"

/*
	dso_linkExternal()
		turn the incoming values into a targeted HREF
 */
char	*dso_linkExternal(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*href = args[0],
			*text = args[1],
			*buf = (char *)malloc(sizeof(char) *
				(strlen(href) + strlen(text) + 100));
		if (strstr(href, "@"))
			sprintf(buf, "<a href=\"mailto:%s\">%s</a>", href, text);
		else
			sprintf(buf, "<a href=\"%s\" target=\"new\">%s</a>", href, text);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(buf);
	} if (num_args == 1) {
		char	*href = args[0],
			*text = args[0],
			*buf = (char *)malloc(sizeof(char) *
				(strlen(href) + strlen(text) + 100));
		if (strstr(href, "@"))
			sprintf(buf, "<a href=\"mailto:%s\">%s</a>", href, text);
		else
			sprintf(buf, "<a href=\"%s\" target=\"new\">%s</a>", href, text);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(buf);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(SYNTAX_linkExternal);
	}
}

/*
	dso_filterProgramArguments()
		filter HTDB-sensitive values
 */
char	*dso_filterProgramArguments(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(filterProgramArguments(in));
}

#define	SYNTAX_truncAt	"usage: \\$\\{truncAt(delimiter, string)\\}"

/*
	dso_truncAt()
		end the string past the delimiting string
 */
char	*dso_truncAt(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*truncAt = args[0],
			*text = strdup(args[1]),
			*c = strcasestr(text, truncAt);
		if (c && *c)
			*c = '\0';
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(text);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(SYNTAX_truncAt);
	}
}

/*
	dso_eval()
		where `token' is a string representation of
		an arithmetic expression, return `token'
		evaluated into a numeric value.
 */
char	*dso_eval(char *in)
{
	double	evaluated;
	char	out[BUF_LIMIT_SMALL + 1];

	if (evaluate(in, &evaluated) == 0) {
		if (evaluated == (int)evaluated)
			snprintf(out, BUF_LIMIT_SMALL, "%d", (int)evaluated);
		else
			snprintf(out, BUF_LIMIT_SMALL, "%g", evaluated);

		out[BUF_LIMIT_SMALL] = '\0';
		return DSO_RETURNS_SUCCESS(out);
	} else
		snprintf(out, BUF_LIMIT_SMALL, "Syntax error within: (%s)", in);

	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS_CHALLENGED(out);
}

char	*dso_min(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 0);

	if (num_args >= 2) {
		char	ret[BUF_LIMIT_SMALL + 1];
		int		i,
				min_index = 0;
		bool_t	isString = BOOL_FALSE;

		for (i = 0 ; BOOL_FALSE == isString && i < num_args ; i++)
			/*
				is this argument non-numeric?
				if so, we'll user string context for all comparisons
			 */
			if (args[i] && *args[i] && !(isdigit((int)args[i][0]) || args[i][0] == '-'))
				isString = BOOL_TRUE;
			
		if (BOOL_FALSE == isString) {
			/*
				numeric context
			 */
			for (i = 1 ; i < num_args ; i++)
				if (atof(args[i]) < atof(args[min_index]))
					min_index = i;
		} else {
			/*
				string context
			 */
			for (i = 1 ; i < num_args ; i++)
				if (strcmp(args[i], args[min_index]) < 0)
					min_index = i;
		}
		sstrncpy(ret, args[min_index]);
		ret[BUF_LIMIT_SMALL] = '\0';
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(in);
	}
}

char	*dso_max(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 0);

	if (num_args >= 2) {
		char	ret[BUF_LIMIT_SMALL + 1];
		int		i,
				min_index = 0;
		bool_t	isString = BOOL_FALSE;

		for (i = 0 ; BOOL_FALSE == isString && i < num_args ; i++)
			/*
				is this argument non-numeric?
				if so, we'll user string context for all comparisons
			 */
			if (args[i] && *args[i] && !(isdigit((int)args[i][0]) || args[i][0] == '-'))
				isString = BOOL_TRUE;
			
		if (BOOL_FALSE == isString) {
			/*
				numeric context
			 */
			for (i = 1 ; i < num_args ; i++)
				if (atof(args[i]) > atof(args[min_index]))
					min_index = i;
		} else {
			/*
				string context
			 */
			for (i = 1 ; i < num_args ; i++)
				if (strcmp(args[i], args[min_index]) > 0)
					min_index = i;
		}
		sstrncpy(ret, args[min_index]);
		ret[BUF_LIMIT_SMALL] = '\0';
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(in);
	}
}

/*
	morse()
		translate and output `str' into morse code.
		god save our souls.
 */
char	*dso_morse(char *in)
{
	spaceData	b;
	static spaceName morse = NULL;

	if (morse == NULL) {
		space_init(morse, HASHTABSIZE_SMALL, htdb_destroyStringTuple);
		space_store(morse, "A", strdup(".-"), 1);
		space_store(morse, "B", strdup("-..."), 1);
		space_store(morse, "C", strdup("-.-."), 1);
		space_store(morse, "D", strdup("-.."), 1);
		space_store(morse, "E", strdup("."), 1);
		space_store(morse, "F", strdup("..-."), 1);
		space_store(morse, "G", strdup("--."), 1);
		space_store(morse, "H", strdup("...."), 1);
		space_store(morse, "I", strdup(".."), 1);
		space_store(morse, "J", strdup(".---"), 1);
		space_store(morse, "K", strdup("-.-"), 1);
		space_store(morse, "L", strdup(".-.."), 1);
		space_store(morse, "M", strdup("--"), 1);
		space_store(morse, "N", strdup("-."), 1);
		space_store(morse, "O", strdup("---"), 1);
		space_store(morse, "P", strdup(".--."), 1);
		space_store(morse, "Q", strdup("--.-"), 1);
		space_store(morse, "R", strdup(".-."), 1);
		space_store(morse, "S", strdup("..."), 1);
		space_store(morse, "T", strdup("-"), 1);
		space_store(morse, "U", strdup("..-"), 1);
		space_store(morse, "V", strdup("...-"), 1);
		space_store(morse, "W", strdup(".--"), 1);
		space_store(morse, "X", strdup("-..-"), 1);
		space_store(morse, "Y", strdup("-.--"), 1);
		space_store(morse, "Z", strdup("--.."), 1);
		space_store(morse, "1", strdup(".----"), 1);
		space_store(morse, "2", strdup("..---"), 1);
		space_store(morse, "3", strdup("...--"), 1);
		space_store(morse, "4", strdup("....-"), 1);
		space_store(morse, "5", strdup("....."), 1);
		space_store(morse, "6", strdup("-...."), 1);
		space_store(morse, "7", strdup("--..."), 1);
		space_store(morse, "8", strdup("---.."), 1);
		space_store(morse, "9", strdup("----."), 1);
		space_store(morse, "0", strdup("-----"), 1);
		space_store(morse, ".", strdup(".-.-.-"), 1);
		space_store(morse, ",", strdup("--..--"), 1);
		space_store(morse, "?", strdup("..--.."), 1);
	}

	data_init(&b);

	while (in && *in) {
		if (isspace((int)*in))
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		else {
			char	lk[2],
				*lookup;
			lk[0] = *in;
			lk[1] = '\0';
			lookup = space_value(morse, lk);
			if (lookup && *lookup) {
				thing_append(&b, lookup, strlen(lookup), BUF_LIMIT_BIG);
				thing_appendChar(&b, ' ', BUF_LIMIT_BIG);
			}
		}
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	demorse()
		translate morse code into cleartext.
		god save our souls.
 */
char	*dso_demorse(char *in)
{
	char	*ptr = strtok(in, "\n\r\t ");
	spaceData	b;
	static spaceName demorse = NULL;

	if (demorse == NULL) {
		space_init(demorse, HASHTABSIZE_SMALL, htdb_destroyStringTuple);
		space_store(demorse, ".-", strdup("A"), 1);
		space_store(demorse, "-...", strdup("B"), 1);
		space_store(demorse, "-.-.", strdup("C"), 1);
		space_store(demorse, "-..", strdup("D"), 1);
		space_store(demorse, ".", strdup("E"), 1);
		space_store(demorse, "..-.", strdup("F"), 1);
		space_store(demorse, "--.", strdup("G"), 1);
		space_store(demorse, "....", strdup("H"), 1);
		space_store(demorse, "..", strdup("I"), 1);
		space_store(demorse, ".---", strdup("J"), 1);
		space_store(demorse, "-.-", strdup("K"), 1);
		space_store(demorse, ".-..", strdup("L"), 1);
		space_store(demorse, "--", strdup("M"), 1);
		space_store(demorse, "-.", strdup("N"), 1);
		space_store(demorse, "---", strdup("O"), 1);
		space_store(demorse, ".--.", strdup("P"), 1);
		space_store(demorse, "--.-", strdup("Q"), 1);
		space_store(demorse, ".-.", strdup("R"), 1);
		space_store(demorse, "...", strdup("S"), 1);
		space_store(demorse, "-", strdup("T"), 1);
		space_store(demorse, "..-", strdup("U"), 1);
		space_store(demorse, "...-", strdup("V"), 1);
		space_store(demorse, ".--", strdup("W"), 1);
		space_store(demorse, "-..-", strdup("X"), 1);
		space_store(demorse, "-.--", strdup("Y"), 1);
		space_store(demorse, "--..", strdup("Z"), 1);
		space_store(demorse, ".----", strdup("1"), 1);
		space_store(demorse, "..---", strdup("2"), 1);
		space_store(demorse, "...--", strdup("3"), 1);
		space_store(demorse, "....-", strdup("4"), 1);
		space_store(demorse, ".....", strdup("5"), 1);
		space_store(demorse, "-....", strdup("6"), 1);
		space_store(demorse, "--...", strdup("7"), 1);
		space_store(demorse, "---..", strdup("8"), 1);
		space_store(demorse, "----.", strdup("9"), 1);
		space_store(demorse, "-----", strdup("0"), 1);
		space_store(demorse, ".-.-.-", strdup("."), 1);
		space_store(demorse, "--..--", strdup(","), 1);
		space_store(demorse, "..--..", strdup("?"), 1);
	}

	data_init(&b);

	do {
		if (ptr) {
			char	*lookup = space_value(demorse, ptr);
			if (lookup && *lookup) {
				thing_appendChar(&b, *lookup, BUF_LIMIT_BIG);
				thing_appendChar(&b, ' ', BUF_LIMIT_BIG);
			}
		}
	} while (ptr && (ptr = strtok(NULL, "\n\r\t ,")));

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_codeify()
		add HTML line breaks and whitespace to the incoming string
 */
char	*dso_codeify(char *in)
{
	spaceData	b;
	data_init(&b);

	while (in && *in) {
		if (*in == '\n')
			thing_append(&b, "<br>\n", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '&')
			thing_append(&b, "&amp;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '\t')
			thing_append(&b, "&nbsp;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}
	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_codeify()
		add HTML line breaks and whitespace to the incoming string
 */
char	*dso_makeHTML(char *in)
{
	spaceData	b;
	data_init(&b);

	while (in && *in) {
		if (*in == '\n')
			thing_append(&b, "<br>\n", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '\t')
			thing_append(&b, "&nbsp;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}
	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_safety()
		poorly-named function.
		turns HTDB-sensitive values into HTML-printable equivalents
 */
char	*dso_safety(char *in)
{
	spaceData	b;
	data_init(&b);

	while (in && *in) {
		if (*in == '#')
			thing_append(&b, "&#35;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '$')
			thing_append(&b, "&#36;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '<')
			thing_append(&b, "&lt;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '>')
			thing_append(&b, "&gt;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '\n')
			thing_append(&b, "<br>\n", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '&')
			thing_append(&b, "&amp;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else if (*in == '\t')
			thing_append(&b, "&nbsp;", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}
	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_describe()

		used to self document HTDB objects
		(to the extent possible)

		behold - the only embedded HTML you will find in htdb source code.
 */
char	*dso_describe(char *in)
{
	spaceData	*b;
	spaceData	bx;
	int	found = 0;

	data_init(&bx);

	if (cfunc_lookup(in)) {
		thing_append(&bx, "<table border=\"1\" width=\"100%\"><tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<td width=\"5%\">name</td><td>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

		thing_append(&bx,
			"<tr><td width=\"5%\">type</td><td>Compiled Function", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		found++;
		thing_append(&bx, "</td></tr></table>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	}
	if ((b = sfunc_lookup(in))) {
		char	tmp[512];
		int	i;
		char	*ptr;
		sfunc_t	*sf = (sfunc_t *)b->value;

		thing_append(&bx, "<table border=\"1\" width=\"100%\"><tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<td width=\"5%\">name</td><td>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

		thing_append(&bx,
			"<tr><td width=\"5%\">type</td><td>Resource File Script Function</td></tr>",
			FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<tr><td width=\"5%%\">usage</td><td>$\\{", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_appendChar(&bx, '(', BUF_LIMIT_BIG);
		for (i = 0 ; i < sf->num_args ; i++) {
			thing_append(&bx, sf->args[i], FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			thing_append(&bx, "=...", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			if (i < sf->num_args - 1)
				thing_append(&bx, ", ", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		}
		thing_append(&bx, ")}</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		sprintf(tmp, "<tr><td width=\"5%%\">bytes</td><td>%ld</td></tr>"
			"<tr><td width=\"5%%\" valign=\"top\">body</td><td><code>",
			(long)strlen(sf->body));
		thing_append(&bx, tmp, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, ptr = dso_safety(sf->body),
			FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		free(ptr);
		found++;
		thing_append(&bx, "</td></tr></table>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	}
	if ((b = static_lookup(in))) {
		char	tmp[512],
				*ptr;
		thing_append(&bx, "<table border=\"1\" width=\"100%\"><tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<td width=\"5%\">name</td><td>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

		thing_append(&bx,
			"<tr><td width=\"5%\">type</td><td>Static Resource File Definition</td></tr>",
			FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		sprintf(tmp,
			"<tr><td width=\"5%%\">bytes</td><td>%ld</td></tr><tr>"
			"<td width=\"5%%\" valign=\"top\">value</td><td><code>",
			(long)b->size);
		thing_append(&bx, tmp, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, ptr = dso_safety(b->value), FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		free(ptr);
		found++;
		thing_append(&bx, "</td></tr></table>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	} else if ((b = htdb_lookup(in))) {
		char	tmp[512],
				*ptr;
		thing_append(&bx, "<table border=\"1\" width=\"100%\"><tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<td width=\"5%\">name</td><td>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

		thing_append(&bx,
			"<tr><td width=\"5%\">type</td><td>Resource File Definition</td></tr>",
			FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		sprintf(tmp,
			"<tr><td width=\"5%%\">bytes</td><td>%ld</td></tr><tr>"
			"<td width=\"5%%\" valign=\"top\">value</td><td><code>",
			(long)b->size);
		thing_append(&bx, tmp, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, ptr = dso_safety(b->value), FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		free(ptr);
		found++;
		thing_append(&bx, "</td></tr></table>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	}
	if (found == 0) {
		thing_append(&bx, "<table border=\"1\" width=\"100%\"><tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "<td width=\"5%\">name</td><td>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, in, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

		thing_append(&bx, "<tr><td width=\"5%\">type</td><td>Undefined", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		thing_append(&bx, "</td></tr></table>", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
	}

	return DSO_RETURNS_SUCCESS_ASIS((char *)bx.value);
}

/*
	dso_rtoi()
		turns roman numerals into integers
 */
char	*dso_rtoi(char *in)
{
	char	*out = (char *)malloc(sizeof(char) * ((strlen(in) * 5) + 1));
	sprintf(out, "%d", rtoi(in));
	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_itor()
		turns integers into roman numerals
 */
char	*dso_itor(char *in)
{
	char	*out = (char *)malloc(sizeof(char) * ((strlen(in) * 5) + 1));

	itor(atoi(in), &out);
	return DSO_RETURNS_SUCCESS_ASIS(out);
}

/*
	dso_literalizeCommas()
		um, literalizes commsa
 */
char	*dso_literalizeCommas(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(literalizeCommas(in));
}

/*
	dso_literalizeQuotes()
		um, literalizes quotes
 */
char	*dso_literalizeQuotes(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(literalizeQuotes(in));
}

/*
	dso_deLiteralizeCommas()
		removes literal marks from commas on input
 */
char	*dso_deLiteralizeCommas(char *in)
{
	spaceData	b;
	data_init(&b);
	while (in && *in) {
		if (*in == '\\' && (in + 1) && *(in + 1) && *(in + 1) == ',')
			;
		else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS((char *)b.value);
}

/*
	dso_igpayatinlay()
		pig-latin encoding
 */
char	*dso_igpayatinlay(char *in)
{
	char	*ptr = strtok(in, "\n\r\t ");
	spaceData	b;

	data_init(&b);

	do {
		if (ptr && *ptr != '<') {
			int		len = strlen(ptr);
			char	*pigish = (char *)malloc(sizeof(ptr) * (len + 3));

			strcpy(pigish, &ptr[1]);
			pigish[len - 1] = ptr[0];
			pigish[len] = '\0';
			strcat(pigish, "ay");

			thing_append(&b, pigish, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			thing_appendChar(&b, ' ', BUF_LIMIT_BIG);
			free(pigish);
		}
	} while (ptr && (ptr = strtok(NULL, "\n\r\t ,")));

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

#define	SENDMAIL_SYNTAX		"usage: \\$\\{sendmail(message)\\}"
#define	SENDMAIL_MISSING	"sendmail undefined in config.htdb"

/*
	dso_sendmail()
		pipelines the input string through whatever resourse
		is defined as `confSendmail'
 */
char	*dso_sendmail(char *source)
{
	if (BOOL_TRUE == amRoot())
		return DSO_RETURNS_SUCCESS_CHALLENGED(ARE_YOU_INSANE);
	else {
		return (sendmail(htdb_getval("confSendmail"), source) == 0) ?
			DSO_RETURNS_SUCCESS_CHALLENGED(SENDMAIL_MISSING) :
			DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	}
}

/*
	dso_ntoa()
		dottted-quad representation of integer input
 */
char	*dso_ntoa(char *in)
{
	unsigned long	t = atof(in),
			x1 = 256,
			x2 = 65536,
			x3 = 16777216,
			o1 = t / x3,
			d1 = t - (o1 * x3),
			o2 = d1 / x2,
			d2 = d1 - (o2 * x2),
			o3 = d2 / x1,
			d3 = d2 - (o3 * x1),
			o4 = d3;
	char	out[BUF_LIMIT_SMALL + 1];
	snprintf(out, BUF_LIMIT_SMALL, "%ld.%ld.%ld.%ld", o1, o2, o3, o4);
	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(out);
}

/*
	dso_aton()
		decoding of dottted-quad representation of network address
 */
char	*dso_aton(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ".", &args, 4);
	unsigned	lout = 0;

	if (num_args == 4)
		lout = (atol(args[0]) * (unsigned)16777216) +
				(atol(args[1]) * (unsigned)65536) +
				(atol(args[2]) * 256) + atol(args[3]);

	htdb_deleteArgs(args, num_args);

	{
		char	out[BUF_LIMIT_SMALL + 1];

		snprintf(out, BUF_LIMIT_SMALL, "%u", lout);
		out[BUF_LIMIT_SMALL] = '\0';
		return DSO_RETURNS_SUCCESS(out);
	}
}

/*
	dso_signed()
		output `+' or `-' on single integer value
 */
char	*dso_signed(char *in)
{
	char	out[BUF_LIMIT_SMALL + 1];

	if (in && *in) {
		if (*in == '-')
			sstrncpy(out, in);
		else
			snprintf(out, BUF_LIMIT_SMALL, "+%s", in);
	} else
		sstrncpy(out, "");
	out[BUF_LIMIT_SMALL] = '\0';
	return DSO_RETURNS_SUCCESS(out);
}

/*
	dso_cleanLines()
		minimizes whitespace in input string
 */
char	*dso_cleanLines(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_cleanLines(in));
}

/*
	dso_imagify()
		sick.  for every printable character in the input,
		spit out an HTML img tag.
 */
char	*dso_imagify(char *in)
{
	int	intag = 0;
	spaceData	b;

	data_init(&b);

	while (in && *in) {
		if (intag && *in == '>') {
			intag = 0;
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		} else if (!intag && *in == '<') {
			intag = 1;
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		} else if (intag || !isalpha((int)*in)) {
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		} else {
			thing_append(&b, "<img border=\"0\" src=\"http://adjective.com/images/alpha/",
				FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
			thing_append(&b, ".gif\"", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			if (isupper((int)*in))
					thing_append(&b, " height=\"14\" width=\"15\" alt=\"", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			else
					thing_append(&b, " height=\"10\" width=\"11\" alt=\"", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
			thing_append(&b, "\">", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
		}
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

/*
	dso_inList()
		returns the substring if it is one of a specified subset
		found within the input string.
 */
char	*dso_inList(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 0),
			found = 0;

	if (num_args > 1) {
		char	*target = args[0];
		int		i;
		for (i = 1 ; !found && i < num_args ; i++)
			if (strcasestr(target, args[i]))
				found = i;
	}

	if (found) {
		char	*foundString = strdup(args[found]);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(foundString);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
	}
}

/*
	dso_rot13()
		super-secret rot13 encryption.  adobe may sue!
 */
char	*dso_rot13(char *in)
{
	spaceData	b;
	data_init(&b);
	while (in && *in) {
		if (*in >= 'A' && *in < 'M')
			thing_appendChar(&b, *in + 13, BUF_LIMIT_BIG);
		else if (*in >= 'a' && *in < 'm')
			thing_appendChar(&b, *in + 13, BUF_LIMIT_BIG);
		else if (*in >= 'N' && *in <= 'Z')
			thing_appendChar(&b, *in - 13, BUF_LIMIT_BIG);
		else if (*in >= 'n' && *in <= 'z')
			thing_appendChar(&b, *in - 13, BUF_LIMIT_BIG);
		else
			thing_appendChar(&b, *in, BUF_LIMIT_BIG);
		in++;
	}

	return DSO_RETURNS_SUCCESS_ASIS((char *)b.value);
}

#define	SYNTAX_MAKEARRAY	"usage: \\$\\{makeArray(target, [separator,] string)\\}"

/*
	dso_makeArray()
		breaks a string into an htdb array object.
		array characteristics are determined by the separator value
 */
char	*dso_makeArray(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3 || num_args == 2) {
		char	*target = args[0],
				*separator = (num_args == 3) ? args[1] : "",
				*string = (num_args == 3) ? args[2] : args[1];
		int	done = 0;

		if (num_args == 3 && strlen(separator) == 0) {
			/*
				empty separator value.
				break on characters.
			 */
			int	index = 0;

			while (string && *string) {
				char	val[2];
				val[0] = *string;
				val[1] = '\0';
				htdb_setarrayval(target, ++index, val);
				string++;
			}
			htdb_setptrint(target, "numResults", index);
			done = 1;
		} else if (num_args == 2) {
			/*
				only 2 arguments.
				break on word boundaries.
			 */
			separator = " ";
		}

		if (!done) {
			char	**n_args;
			int		n_num_args = htdb_splitArgs(string, separator, &n_args, 0),
					i;

			for (i = 0 ; i < n_num_args ; i++)
				htdb_setarrayval(target, i + 1, n_args[i]);

			htdb_setptrint(target, "numResults", n_num_args);

			htdb_deleteArgs(n_args, n_num_args);
		}

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS("");
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_MAKEARRAY);
	}
}

char	*dso_forEach(char *in)
{
	int		i;
	char	**args,
			forEachPrefix[BUF_LIMIT_BIG + 1];
	int		num_args = htdb_splitArgs(in, ",", &args, 0);

	snprintf(forEachPrefix, BUF_LIMIT_BIG, args[0]);
	forEachPrefix[BUF_LIMIT_BIG] = '\0';
	for (i = 1; i < num_args; i++) {
		htdb_setobjval(forEachPrefix, i, "val", args[i]);
	}
	htdb_setptrint(forEachPrefix, "numResults", i);
	htdb_deleteArgs(args, num_args);
	return DSO_RETURNS_SUCCESS("");
}

char	*dso_stripWhitespace(char *in)
{
	char *work = strdup(in);
	htdb_stripWhitespace(work);
	return DSO_RETURNS_SUCCESS_ASIS(work);
}

#define	SYNTAX_TERNARY	"usage: \\$\\{ternary((expression), true, false)\\}"

char	*dso_ternary(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);
	char	*expression = "",
			*caseTrue = "",
			*caseFalse = "",
			*out = "";

	if (num_args == 2) {
		expression = strdup(args[0]);
		caseTrue = args[1];
	} else if (num_args == 3) {
		expression = strdup(args[0]);
		caseTrue = args[1];
		caseFalse = args[2];
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_TERNARY);
	}

	if (BOOL_TRUE == live_eval_expr(expression))
		out = htdb_expand(caseTrue);
	else
		out = htdb_expand(caseFalse);

	free(expression);

	htdb_deleteArgs(args, num_args);

	return DSO_RETURNS_SUCCESS(out);
}

#define	SYNTAX_IMAGESCALEANDSTORE	"usage: \\$\\{imageScaleAndStore(table, dir, file, name, width, height)\\}"

char	*dso_imageScaleAndStore(char *in)
{
	bool_t	status = BOOL_FALSE;
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 6);

	if (num_args == 6) {
		char	*table = args[0],
				*dir = args[1],
				*file = args[2],
				*name = args[3];
		int_t	user_id = htdb_getint("user->user_id");
		int		width = atoi(args[4]),
				height = atoi(args[5]);
		status = imageScaleAndStore(table, user_id, name, dir, file, width, height);
	}

	htdb_deleteArgs(args, num_args);

	if (BOOL_TRUE == status)
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_IMAGESCALEANDSTORE);

}

#define	SYNTAX_MAPSTR	"usage: \\$\\{mapStr(pattern1, pattern2, string)\\}"

char	*dso_mapStr(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	*pattern1 = args[0],
				*pattern2 = args[1],
				*string = args[2],
				*str = htdb_mapStr(pattern1, pattern2, string);

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(str);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_MAPSTR);
	}
}

char	*dso_summarizeText(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 5);

	if (num_args == 5) {
		int_t	lenBeforeText= atoi(args[0]),
				lenAfterText = atoi(args[1]);
		char	*term = args[2],
				*hilited = args[3],
				*text = args[4],
				*ptr = strcasestr(text, term);

		if (lenBeforeText < 0 && lenBeforeText > 100)
			lenBeforeText = 10;

		if (lenAfterText < 0 && lenAfterText > 100)
			lenAfterText = 20;

		if (ptr && *ptr) {
			spaceData   b;
			int		pos = 0;
			char	*ret;

			data_init(&b);

			/*
				calculate how much "prior" text to add
			 */
			if ((ptr - lenBeforeText) <= text) {
				pos = 0;
			} else {
				thing_append(&b, "...", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
				pos = ptr - text - lenBeforeText;
			}
			/*
				add "prior" text
			 */
			while (&text[pos] < ptr) {
				thing_appendChar(&b, text[pos], BUF_LIMIT_BIG);
				pos++;
			}
			/*
				add the thing
			 */
			thing_append(&b, hilited, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);

			pos += strlen(term);

			{
				int		num = 0;

				for (pos = pos ; num++ < lenAfterText && text[pos] ; pos++) {
					thing_appendChar(&b, text[pos], BUF_LIMIT_BIG);
				}
				if (text[pos])
					thing_append(&b, "...", FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			}

			ret = strdup((char *)b.value);
			data_destroy(&b);

			htdb_deleteArgs(args, num_args);
			return DSO_RETURNS_SUCCESS_ASIS(ret);
		}
	}
	htdb_deleteArgs(args, num_args);
	return DSO_RETURNS_SUCCESS("");
}

char	*dso_replaceText(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	*pattern1 = args[0],
				*pattern2 = args[1],
				*input = args[2];
		bool_t	inTag = BOOL_FALSE;

		spaceData   b,
					textBuffer;

		data_init(&b);
		data_init(&textBuffer);

		while (input && *input) {
			if (*input == '>' && BOOL_TRUE == inTag) {
				/*
					end of the tag
				 */
				inTag = BOOL_FALSE;

				/*
					finish off the tag buffer value
				 */
				thing_appendChar(&b, *input, BUF_LIMIT_BIG);

			} else if (*input == '<') {
				/*
					do substitution on the text buffer
				 */
				if ((char *)textBuffer.value && *(char *)textBuffer.value) {
					char	*mapped = htdb_mapStr(pattern1, pattern2, (char *)textBuffer.value);
					thing_append(&b, mapped, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
					free(mapped);
					data_destroy(&textBuffer);
					data_init(&textBuffer);
				}
				/*
					into tag
				 */
				inTag = BOOL_TRUE;
				thing_appendChar(&b, *input, BUF_LIMIT_BIG);
			} else if (inTag == BOOL_FALSE) {
				/*
					add text not in tag to buffer
				 */
				thing_appendChar(&textBuffer, *input, BUF_LIMIT_BIG);
			} else {
				thing_appendChar(&b, *input, BUF_LIMIT_BIG);
			}
			input++;
		}
		if ((char *)textBuffer.value &&
			*(char *)textBuffer.value) {
			char	*mapped = htdb_mapStr(pattern1, pattern2, (char *)textBuffer.value);
			thing_append(&b, mapped, FUNC_CALCULATES_SIZE, BUF_LIMIT_BIG);
			free(mapped);
		}
		htdb_deleteArgs(args, num_args);
		data_destroy(&textBuffer);
		{
			char	*ret = strdup((char *)b.value);
			data_destroy(&b);
			return DSO_RETURNS_SUCCESS_ASIS(ret);
		}
	}
	htdb_deleteArgs(args, num_args);
	return DSO_RETURNS_SUCCESS_CHALLENGED("");
}

char	*dso_stripTags(char *in)
{
	int	intag = 0,
		inexp = 0;
	spaceData	b;
	char	*filtered = htdb_unfilter_unsafe(in),
			*expanded = htdb_expand(filtered),
			*hold = &expanded[0];

	free(filtered);

	data_init(&b);

	while (expanded && *expanded) {
		if (intag && *expanded == '>') {
			intag = 0;
		} else if (!intag && *expanded == '<') {
			intag = 1;
		} else if (inexp > 0 && *expanded == '}') {
			inexp--;
		} else if (*expanded == '$' && (expanded + 1) && *(expanded + 1) && *(expanded + 1) == '{') {
			inexp++;
		} else if (0 == intag && 0 == inexp) {
			thing_appendChar(&b, *expanded, BUF_LIMIT_BIG);
		}
		expanded++;
	}
	free(hold);

	return DSO_RETURNS_SUCCESS_ASIS(b.value);
}

#define	SYNTAX_FETCHVERBAGE		"usage: \\$\\{fetchVerbage(prefix, lookup)\\}"

char	*dso_fetchVerbage(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*prefix = args[0],
				*lookup = args[1];
		int_t	num = htdb_fetchVerbage(prefix, lookup);

		htdb_deleteArgs(args, num_args);

		if (num > 0)
			return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
		else
			return DSO_RETURNS_SUCCESS_CHALLENGED(GENERIC_FAILURE_STRING);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_FETCHVERBAGE);
	}
}
/*
	date/time formatting
 */
char	*dso_parseDate(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseDate(in));
}

char	*dso_parseDateTime(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseDateTime(in));
}

char	*dso_parseTime(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseTime(in));
}

char	*dso_parseMonthNum(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseMonthNum(in));
}

char	*dso_parseMon(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseMon(in));
}

char	*dso_parseMonth(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseMonth(in));
}

char	*dso_parseDay(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseDay(in));
}

char	*dso_parseYear(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseYear(in));
}

char	*dso_parseHour(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseHour(in));
}

char	*dso_parseMinute(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseMinute(in));
}

char	*dso_parseSecond(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseSecond(in));
}

char	*dso_parseAMPM(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseAMPM(in));
}

char	*dso_parseWeekday(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseWeekday(in));
}

char	*dso_parseUnix(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_parseUnix(in));
}

char	*dso_date(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(htdb_date(in));
}

/*	
	routines for entering and exiting secure server mode; 
	assumes the same server name,
	ie http://www.htdb.org and https://www.htdb.org
*/

char	*dso_enterSecure(char *in)
{
	char buf[BUF_LIMIT_BIG + 1]; 
	
	if (strncasecmp(in, "http", 4) != 0)
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
	
	if (tolower(in[4]) == 's')	/* already in secure mode */
		return DSO_RETURNS_SUCCESS(in);

	snprintf(buf, BUF_LIMIT_BIG, "https%s", in + 4);
	return DSO_RETURNS_SUCCESS(buf);
}

char	*dso_exitSecure(char *in)
{
	char buf[BUF_LIMIT_BIG + 1]; 
	
	if (strncasecmp(in, "http", 4) != 0)
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
	
	if (tolower(in[4]) == ':')	/* already in insecure mode */
		return DSO_RETURNS_SUCCESS(in);

	snprintf(buf, BUF_LIMIT_BIG, "http%s", in + 5);
	return DSO_RETURNS_SUCCESS(buf);
}

char	*dso_braille(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(braille(in));
}

char	*dso_ordinalSuffix(char *in)
{
	return DSO_RETURNS_SUCCESS(htdb_ordinalSuffix(atoi(in)));
}

char	*dso_isUnconfirmed(char *in)
{
	if (BOOL_TRUE == isUnconfirmed())
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
}

char	*dso_isConfirmed(char *in)
{
	if (BOOL_TRUE == isConfirmed())
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
}

char	*dso_isUser(char *in)
{
	if (BOOL_TRUE == isUser())
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
}

char	*dso_isAdmin(char *in)
{
	if (BOOL_TRUE == isAdmin())
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	else
		return DSO_RETURNS_SUCCESS_CHALLENGED("");
}

char	*parseHometown(char *in)
{
	char	*city = "",
			*state = "",
			*country = "",
			*state_select = "",
			*state_writein = "",
			*country_select = "",
			*country_writein = "",
			out[BUF_LIMIT_SMALL + 1],
			**args;
	int		num_args = htdb_splitArgs(in, ":", &args, 5);

	if (num_args > 0)
		city = args[0];
	if (num_args > 1)
		state_select = args[1];
	if (num_args > 2)
		state_writein = args[2];
	if (num_args > 3)
		country_select = args[3];
	if (num_args > 4)
		country_writein = args[4];

	state = (state_select && *state_select && *state_select != '-' &&
		*state_select != '*') ? state_select :
		(state_writein && *state_writein && *state_writein != '-' &&
		*state_writein != '*') ? state_writein : "";
	country = (country_select && *country_select && *country_select != '-' &&
		*country_select != '*') ? country_select :
		(country_writein && *country_writein && *country_writein != '-' &&
		*country_writein != '*') ? country_writein : "";

	if (strcasecmp(country, "USA") == 0)
		country = "";

	if (state && *state)
		if (strcasecmp(city, state) == 0)
			state = "";
	if (state && *state)
		if (country && *country)
			if (strcasecmp(state, country) == 0)
				state = "";

	snprintf(out, BUF_LIMIT_SMALL, "%s%s%s%s%s",
		city,
		(city && *city && state && *state) ? ", " : "", state,
		(((city && *city) || (state && *state)) && country && *country) ? ", " : "", country);

	htdb_deleteArgs(args, num_args);

	return pretty(out);
}

/*
	(city:state_select:state_writein:country_select:country_writein)
 */
char	*dso_smartHometown(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(parseHometown(in));
}

/*
	(trunc, lookup_prefix)
 */
char	*dso_easyHometown(char *in)
{
	char	buf[BUF_LIMIT_BIG + 1] = { "" },
			**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2),
			trunc = atoi(num_args > 0 ? args[0] : "");

	strcpy(buf, "");

	if (num_args == 2)
		snprintf(buf, BUF_LIMIT_BIG, "%s:%s:%s:%s:%s",
			htdb_getptrval(args[1], "city"),
			htdb_getobjval("state",
				myIndexOf("state", "state_id", htdb_getptrval(args[1], "state_id")), "name"),
			htdb_getptrval(args[1], "statewritein"),
			htdb_getobjval("country",
				myIndexOf("country", "country_id", htdb_getptrval(args[1], "country_id")), "name"),
			htdb_getptrval(args[1], "country")); 

	{
		char	*prettified = parseHometown(buf);

		if (!*prettified) {
			/*
				no result?  use "unspecified".  lame. 
			 */
			free(prettified);
			prettified = strdup("Unspecified");
		}

		if (trunc && trunc < strlen(prettified)) {
			prettified[trunc] = '\0';
			prettified = (char *)realloc(prettified, sizeof(char) * (trunc + 5));
			strcat(prettified, "...");
		}

		htdb_deleteArgs(args, num_args);

		return DSO_RETURNS_SUCCESS_ASIS(prettified);
	}
}

#define	SYNTAX_ttgDetail	"usage: ttgDetail(resource, domain_id, precision, bars, limit)"

char	*dso_ttgDetail(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 5);

	if (num_args >= 4) {
		char	*resource = args[0];
		int		domain_id = atoi(args[1]),
				precision = atoi(args[2]),
				bars = atoi(args[3]),
				limit = (num_args == 5) ? atoi(args[4]) : 0;
		char	*ret = ttgDetail(resource, domain_id, precision, bars, limit);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(ret);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_ttgDetail);
	}
}

#define	SYNTAX_sort	"usage: sort(obj_name, sort_field)"

char	*dso_sort(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		htdb_objsort(args[0], args[1]);
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS(GENERIC_SUCCESS_STRING);
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_ttgDetail);
	}
}

#define	SYNTAX_fopen	"usage: \\$\\{dso_fopen(handle, path, mode)\\}"
char    *dso_fopen(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 3);

	if (num_args == 3) {
		char	*prefix = args[0],
				*path = args[1],
				*mode = args[2];
		spaceData	*b = file_lookup(prefix);

		if (b == (spaceData *)NULL) {
			filep_t	*fpt = (filep_t *)malloc(sizeof(filep_t));
			/*
				new file request
			 */
			fpt->path = strdup(path);
			fpt->mode = strdup(mode);
			(FILE *)fpt->fp = (FILE *)fopen(fpt->path, fpt->mode);
			file_store(prefix, (void *)fpt);
		} else {
			/*
				file's already open
			 */
		}

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS("");
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fopen);
	}
}

#define	SYNTAX_fclose	"usage: \\$\\{dso_fclose(handle)\\}"
char    *dso_fclose(char *in)
{
	char	*prefix = in;

	if (prefix && *prefix) {
		spaceData	*b = file_lookup(prefix);

		if (b == (spaceData *)NULL) {
			return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fclose);
		} else {
			/*
				close the file, clean up.
			 */
			htdb_destroyFilePointer(b->value);
			b->value = NULL;
			free(b->key);
			b->key = strdup("");
			return DSO_RETURNS_SUCCESS("");
		}

	} else {
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fclose);
	}
}

#define	SYNTAX_fwrite	"usage: \\$\\{dso_fwrite(handle, string)\\}"
char    *dso_fwrite(char *in)
{
	char	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*prefix = args[0],
				*string = args[1];
		spaceData	*b = file_lookup(prefix);

		if (b == (spaceData *)NULL) {
			htdb_deleteArgs(args, num_args);
			return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fwrite);
		} else {
			filep_t	*fpt = (filep_t *)b->value;
			fprintf((FILE *)fpt->fp, string);
		}

		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS("");
	} else {
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_fwrite);
	}
}

#define	SYNTAX_AVG	"usage: \\$\\{avg(value [, value [,value...]])\\}"

char	*dso_avg(char *in)
{
	char	**args;
	int		i,
			num_args = htdb_splitArgs(in, ",", &args, 0);
	double	sum = 0.0,
			avg = 0.0;
	char	ret[BUF_LIMIT_SMALL + 1];
	bool_t	allInt = BOOL_TRUE;

	for (i = 0 ; i < num_args ; i++) {
		sum += atof(args[i]);
		if (strchr(args[i], '.'))
			allInt = BOOL_FALSE;
	}
	htdb_deleteArgs(args, num_args);

	if (num_args > 0)
		avg = sum / num_args;

	snprintf(ret, BUF_LIMIT_SMALL, (BOOL_FALSE == allInt ? "%.2f" : "%.0f"), avg);

	return DSO_RETURNS_SUCCESS(ret);
}

char	*dso_unescape(char *in)
{
	char	*out = strdup(in);
	uncgi_url_unescape(out);
	return DSO_RETURNS_SUCCESS_ASIS(out);
}

char	*dso_escape(char *in)
{
	return DSO_RETURNS_SUCCESS_ASIS(url_escape(in));
}

#define	SYNTAX_googleSearch	"usage: googleSearch(key, query)"
char	*dso_googleSearch(char *in)
{
	char 	**args;
	int		num_args = htdb_splitArgs(in, ",", &args, 2);

	if (num_args == 2) {
		char	*results = doGoogleSearch(args[0], args[1], 0, 10, BOOL_FALSE, "", BOOL_FALSE, "");
		htdb_deleteArgs(args, num_args);
		return DSO_RETURNS_SUCCESS_ASIS(results);
	} else {
		htdb_deleteArgs(args, num_args);
		return	DSO_RETURNS_SUCCESS_CHALLENGED(SYNTAX_googleSearch);
	}
}

char	*dso_ttg(char *in)
{
	return htdb_ttg(in);
}

char	*dso_md5base64(char *in)
{
	char b64_md5[320];
	{
		char from_md5[32];
		memset(from_md5, 0, 32);
		memset(b64_md5, 0, 32);
		md5_digest(in, from_md5, 0, strlen(in));
		base64_encode(from_md5, 16, b64_md5);
	}
	return DSO_RETURNS_SUCCESS(b64_md5);
}
